

#include "stdafx.h"


#include "os.h"
#include "flex.h"
#include "bbc.h"
#include "wimp.h"

#include "etc.h"

#include "deb.h"
#include "xmath.h"
#include "colour.h"

#include "veclib.h"



#define VCHUNK 512

static int linecolour;
static int fillcolour;

#define VEND       0x0
#define VMOVE      0x1
#define VDRAW      0x2
#define VCURVE     0x3
#define VCOLOUR    0x4
#define VSTARTPATH 0x5
#define VCLOSEPATH 0x6


typedef struct vecstr
{
 int   index;
 int   size;
 int * data;
 int   xlo;
 int   xhi;
 int   ylo;
 int   yhi;
 int   vecnobounds;

} vecstr;



static vecstr vec1;



os_error * vecgetindex(vechandle handle,int * vindex)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

 v=(vecstr*)handle;

 *vindex=v->index;

 return(err);
}



int vecsize(vechandle handle)
{
 vecstr   * v;


 v=(vecstr*)handle;

 return(v->index);
}



os_error * vecclose(vechandle handle)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

 v=(vecstr*)handle;

 if(v->index>=v->size)
 {
  err=flex_extend((flex_ptr)&v->data,(v->size+VCHUNK)*sizeof(int));
  if(!err) v->size+=VCHUNK;
 }

 if(!err)
 {
  v->data[v->index++]=VEND;
 }

 return(err);
}


os_error * vecstartpath(vechandle handle)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

 v=(vecstr*)handle;

 if(v->index>=v->size)
 {
  err=flex_extend((flex_ptr)&v->data,(v->size+VCHUNK)*sizeof(int));
  if(!err) v->size+=VCHUNK;
 }

 if(!err)
 {
  v->data[v->index++]=VSTARTPATH;
 }

 return(err);
}



os_error * vecclosepath(vechandle handle)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

 v=(vecstr*)handle;

 if(v->index>=v->size)
 {
  err=flex_extend((flex_ptr)&v->data,(v->size+VCHUNK)*sizeof(int));
  if(!err) v->size+=VCHUNK;
 }

 if(!err)
 {
  v->data[v->index++]=VCLOSEPATH;
 }

 return(err);
}



void veclimits(int x,int y,vechandle handle)
{
 vecstr   * v;

 v=(vecstr*)handle;

 if(v->vecnobounds)
 {
  v->xlo=v->xhi=x;
  v->ylo=v->yhi=y;
  v->vecnobounds=0;
 }
 else
 {
  if(x<v->xlo) {v->xlo=x;}
  if(x>v->xhi) {v->xhi=x;}
  if(y<v->ylo) {v->ylo=y;}
  if(y>v->yhi) {v->yhi=y;}
 }
}






os_error * vecmove(int x, int y,vechandle handle)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

// cprintf("move %d %d",x,y);

 v=(vecstr*)handle;

// veclimits(x,y,handle);

 if((v->index+3)>=v->size)
 {
  err=flex_extend((flex_ptr)&v->data,(v->size+VCHUNK)*sizeof(int));
  if(!err) v->size+=VCHUNK;
 }

 if(!err)
 {
  v->data[v->index++]=VMOVE;
  v->data[v->index++]=x;
  v->data[v->index++]=y;
 }

 return(err);
}



os_error * vecdraw(int x,int y,vechandle handle)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

// cprintf("draw %d %d",x,y);

 v=(vecstr*)handle;

// veclimits(x,y,handle);

 if((v->index+3)>=v->size)
 {
  err=flex_extend((flex_ptr)&v->data,(v->size+VCHUNK)*sizeof(int));
  if(!err) v->size+=VCHUNK;
 }

 if(!err)
 {
  v->data[v->index++]=VDRAW;
  v->data[v->index++]=x;
  v->data[v->index++]=y;
 }

 return(err);
}





os_error * veccurve(int x1,int y1,int x2,int y2,int x3,int y3,vechandle handle)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

// cprintf("curve %d %d %d %d %d %d",x1,y1,x2,y2,x3,y3);

 v=(vecstr*)handle;

// veclimits(x1,y1,handle);
// veclimits(x2,y2,handle);
// veclimits(x3,y3,handle);

 if((v->index+7)>=v->size)
 {
  err=flex_extend((flex_ptr)&v->data,(v->size+VCHUNK)*sizeof(int));
  if(!err) v->size+=VCHUNK;
 }

 if(!err)
 {
  v->data[v->index++]=VCURVE;
  v->data[v->index++]=x1;
  v->data[v->index++]=y1;
  v->data[v->index++]=x2;
  v->data[v->index++]=y2;
  v->data[v->index++]=x3;
  v->data[v->index++]=y3;
 }

 return(err);
}


os_error * veccolour(int linecol,int fillcol,vechandle handle)
{
 os_error * err;
 vecstr   * v;

 err=NULL;

 v=(vecstr*)handle;

 if((v->index+3)>=v->size)
 {
  err=flex_extend((flex_ptr)&v->data,(v->size+VCHUNK)*sizeof(int));
  if(!err) v->size+=VCHUNK;
 }

 if(!err)
 {
  v->data[v->index++]=VCOLOUR;
  v->data[v->index++]=linecol;
  v->data[v->index++]=fillcol;
 }

 return(err);
}






os_error * vecopen(vechandle * handle)
{
 os_error * err;

 *handle=(vechandle)&vec1;

 vec1.index=0;
 vec1.size=VCHUNK;
 vec1.vecnobounds=1;

 if(!vec1.data) err=flex_alloc((flex_ptr)&vec1.data,VCHUNK*sizeof(int));
 else           err=flex_extend((flex_ptr)&vec1.data,VCHUNK*sizeof(int));

 return(err);
}




os_error * vecscrap(vechandle * handle)
{
 os_error * err;

 err=flex_free((flex_ptr)&((vecstr*)handle)->data);
 if(!err)
 {
  *handle=NULL;
 }

 return(err);
}




os_error * vecpaint(int ox,int oy,int w,int h,vechandle vhandle,int index)
{
 vecstr   * v;
 int      * p;
 int      * lt;
 int        x;
 int        y;
 int        maxw;
 int        maxh;
 int        xlo;
 int        ylo;
 int        x0;
 int        y0;
 int        x1;
 int        y1;
 int        x2;
 int        y2;

 v=(vecstr*)vhandle;

 p=v->data+index;
 lt=v->data+v->index;
 maxw=v->xhi-v->xlo;
 maxh=v->yhi-v->ylo;
 xlo=v->xlo;
 ylo=v->ylo;


 setgcol(0);
 bbc_setpolyfillmode(0);

 while(p<lt)
 {
  switch(*p++)
  {
   case       VEND:
                   p=lt;
                   break;

   case      VMOVE:
                   x=*p++;
                   y=*p++;
                   bbc_move(ox+scale(x-xlo,w,maxw),oy+scale(y-ylo,h,maxh));
                //   cprintf("vmove %d %d",scale(x-xlo,w,maxw),scale(y-ylo,h,maxh));
                   break;

   case      VDRAW:
                   x=*p++;
                   y=*p++;
                   bbc_draw(ox+scale(x-xlo,w,maxw),oy+scale(y-ylo,h,maxh));
                //   cprintf("h %d w %d maxw %d maxh %d",h,w,maxw,maxh);
                   cprintf("vdraw %d %d",ox+scale(x-xlo,w,maxw),ox+scale(y-ylo,h,maxh));
                   break;

   case     VCURVE:
                   x0=ox+scale((*p++)-xlo,w,maxw);
                   y0=oy+scale((*p++)-ylo,h,maxh);
                   x1=ox+scale((*p++)-xlo,w,maxw);
                   y1=oy+scale((*p++)-ylo,h,maxh);
                   x2=ox+scale((*p++)-xlo,w,maxw);
                   y2=oy+scale((*p++)-ylo,h,maxh);
                   bbc_polybezierto(x0,y0,x1,y1,x2,y2);
                   break;

   case    VCOLOUR:
                   x=*p++;
                   y=*p++;
                   setgcol(x);
                   break;


   case VSTARTPATH:
                   bbc_startpath();
                   break;


   case VCLOSEPATH:
                   bbc_closefigure();
                   bbc_endpath();
                   bbc_strokefillpath();
                   //AbortPath(currenthdc);
                   break;
  }
 }

 return(NULL);
}